/********************************************************************
 *     Copyright (c) 2000 by Hewlett-Packard, All Rights Reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *     GNU Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *     MA 02111-1307 USA
 *
 ********************************************************************/

package hp.infrared;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileDescriptor;

/** 
 * This class implements a server socket over Infrared medium. 
 * 
 * @see IrSocket
 */

public class IrServerSocket
{
  private IrSocketImpl impl = null;

  static {
	System.loadLibrary("irsock");
  }
  
  /** 
   * Creates an infrared server socket on a specified port.
   * @param port the port number
   * 
   * @exception IOException
   * 	if an I/O error occurs while creating the server socket
   */

  public IrServerSocket(int port) throws IOException
  {
	this(port, 0);	
  }

  /** 
   * Creates an infrared server socket on a specified port with the specified
   * backlog.
   * @param port the port number
   * @backlog backlog the backlog
   *  
   * @exception IOException
   * 	if an I/O error occurs while creating the server socket
   */

  public IrServerSocket(int port, int backlog) throws IOException
  {
    impl = new IrSocketImpl();
    impl.create(true);

    try{
      impl.bind(null,port);
      impl.listen(backlog);
    }catch (IOException e){
      impl.close();
      throw e;
    }
  }

  /** 
   * Creates an infrared server socket attached to the specified servicename.
   * 
   * @param servicename name of the service
   *  
   * @exception IOException
   * 	if an I/O error occurs while creating the server socket
   */

  public IrServerSocket(String servicename) throws IOException
  {
	this(servicename, 0 );	
  }

  /** 
   * Creates an infrared server socket attached to the specified servicename 
   * and the specified backlog.
   * 
   * @param servicename name of the service
   * @backlog backlog the backlog
   *  
   * @exception IOException
   * 	if an I/O error occurs while creating the server socket
   */

  public IrServerSocket(String servicename, int backlog) throws IOException
  {
    impl = new IrSocketImpl();
    impl.create(true);

    try{
      impl.bind(null,servicename);
      impl.listen(backlog);
    }catch (IOException e){
      impl.close();
      throw e;
    }
  }
  
  /** 
   * Listens for a connection to be made to this socket and accepts it.
   * 
   * @return An IrSocket for the connection.
   * @exception IOException
   * 	if an I/O error occurs while waiting for the connection.
   */

  public IrSocket accept() throws IOException
  {
    IrSocket ret;
    ret = new IrSocket();
    impl.accept(ret.getImpl());
    return ret;
  }

 /**
  * Closes this socket.
  *
  * @exception IOException
  * if I/O error occurs when closing the socket.
  */

  public void close() throws IOException
  {
	impl.close();
  }

}

