/********************************************************************
 *     Copyright (c) 2000 by Hewlett-Packard, All Rights Reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *     GNU Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public 
 *     License along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *     MA 02111-1307 USA
 *
 ********************************************************************/

package hp.infrared;

import java.io.FileDescriptor;
import java.io.IOException;
import java.net.SocketException;
import java.net.BindException;

/**
 * This class implements a connectionless broadcast socket using Ultra protocol.
 * @see UltraPacket
 */

public class UltraSocket
{
  static {
	System.loadLibrary("irsock");
	}

  private final static int MAX_LSAP_SEL = 127;
   
  private UltraSocketImpl impl;

  /**
   * Constructs an ultra socket with the specified protocol id
   * @param pid the ultra protocol id
   * @exception SocketException
   *	if I/O error occurs
   */

  public UltraSocket(int pid) throws SocketException
    {
      if ((pid < 0) || (pid > MAX_LSAP_SEL))
	{
	  throw new IllegalArgumentException("PID out of range:" + pid);
	}

      impl = new UltraSocketImpl();

      impl.create();

      try
	{
	  impl.bind(pid);
	}
      catch (BindException e)
	{
	  impl.close();
	  throw new SocketException();
	}
    }

  /**
   * Closes this ultra socket.
   */

  public void close()
    {
      impl.close();
    }

  /** 
   * Receives an ultra packet from this socket. When this method returns, 
   * the UltraPacket's buffer is filled with the data received. 
   * This method blocks until a packet is received. The length field of the 
   * ultra packet object contains the length of the received message. If the    
   * message is longer than the buffer length, the message is truncated. 
   *
   * @param pkt the packet received
   * @exception IOException
   * 	if any I/O error occurs
   */

  public void receive(UltraPacket pkt) throws IOException
    {
      if (pkt == null)
	{
	  throw new NullPointerException();
	}
      else
	{
	  if (pkt.getData() == null)
	    throw new NullPointerException();
	  else if (pkt.getLength() < 0)
	    throw new IOException("Invalid argument");
	  else if (pkt.getLength() > 0)
	    impl.receive(pkt);
	}
    }

  /**
   * Sends an ultra packet from this socket. The UltraPacket includes 
   * information indicating the data to be sent and its length.
   * @param pkt the ultra packet
   * @exception IOException
   * 	if an I/O error occurs
   */

  public void send(UltraPacket pkt) throws IOException
    {
      if (pkt == null)
	{
	  throw new NullPointerException();
	}
      else
	{
	  if (pkt.getData() == null)
	    throw new NullPointerException();
	  else if (pkt.getLength() < 0)
	    throw new IOException("Invalid argument");
	  else
	    impl.send(pkt);
	}
    }
}
