/********************************************************************
 *     Copyright (c) 2000 by Hewlett-Packard, All Rights Reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; either
 *     version 2.1 of the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *     GNU Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public 
 *     License along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *     MA 02111-1307 USA
 *
 ********************************************************************/

#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <linux/irda.h>

#include "hp_infrared_IrSocketImpl.h"

static int fdValue(JNIEnv * env, jobject fdobj)
{
  return (*env)->GetIntField(env,fdobj, (*env)->GetFieldID(env,(*env)->GetObjectClass(env,fdobj), "fd", "I"));
}

static int doAccept(int s, int timeout, int* cfd, struct sockaddr* caddr, socklen_t* caddrLen, const char** errmsg)
{
  fd_set readset, errorset;
  struct timeval timeoutval;
  struct timeval* timeoutptr;
  int ret;

  do {
  FD_ZERO(&readset);
  FD_ZERO(&errorset);
  FD_SET(s, &readset);
  FD_SET(s, &errorset);
  if (timeout == 0){
    timeoutptr = 0;
  }else{
    timeoutptr = &timeoutval;
    timeoutptr->tv_sec = timeout / 1000;
    timeoutptr->tv_usec = (timeout % 1000) * 1000;
  }
  errno = 0;
  ret = select(s+1, &readset, 0, &errorset, timeoutptr);
  } while ((ret== -1) && (errno==EINTR)); 

  if (ret > 0){
    *cfd = accept(s, caddr, caddrLen);
    if (*cfd == -1){
      ret = -1;
      *errmsg = strerror(errno);
    }else{
      ret = 0;
    }
  }else if (ret == 0){
    // timeout
    ret = -2;
    *errmsg = "Read timeout";
  }else if (ret == -1){
    // something wrong when select
    ret = -1; 
    *errmsg = strerror(errno);
  }

  return ret;
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketCreate
 * Signature: (Z)Ljava/io/FileDescriptor;
 */
JNIEXPORT jobject JNICALL Java_hp_infrared_IrSocketImpl_socketCreate
(JNIEnv *env, jclass cls, jboolean isStream)
{
  int type,fd;
  jobject res;
  if (isStream)
    type = SOCK_STREAM;
  else {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Dgram sockets not supported for Ir");
 } 
  fd = socket(AF_IRDA, type, 0);

  if (fd != -1){
    jclass clazz;
    jfieldID id;
    clazz = (*env)->FindClass(env,"java/io/FileDescriptor");
    id = (*env)->GetFieldID(env,clazz, "fd", "I");
    res = (*env)->NewObject(env,clazz, (*env)->GetMethodID(env,clazz, "<init>", "()V"));
    (*env)->SetIntField(env,res, id, fd);
  }else{
    res = 0;
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
  }

  return res;
}
  

/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketBind
 * Signature: (Ljava/io/FileDescriptor;[BI)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketBind
(JNIEnv *env, jclass cls, jobject fdobj, jbyteArray jip, jint port)
{
  int fd,res;
  struct sockaddr_irda addr;
  
  fd = fdValue(env, fdobj);
  memset(&addr, 0, sizeof addr);
  addr.sir_family = AF_IRDA;
  sprintf(addr.sir_name,"Service-%d",port);
  addr.sir_lsap_sel = LSAP_ANY;

  res = bind(fd, (struct sockaddr*)&addr, sizeof addr);

  if ( res == -1){
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
  }

}
    
/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketBindObject
 * Signature: (Ljava/io/FileDescriptor;[BLjava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketBindObject
(JNIEnv *env, jclass cls, jobject fdobj, jbyteArray jip, jstring IASObject)
{
  int fd,res;
  struct sockaddr_irda addr;
  const char *objectname;
 
  fd = fdValue(env, fdobj);
  memset(&addr, 0, sizeof addr);
  addr.sir_family = AF_IRDA;
  objectname = (*env)->GetStringUTFChars(env,IASObject,NULL);
  strcpy(addr.sir_name,objectname);
  addr.sir_lsap_sel = LSAP_ANY;

  res = bind(fd, (struct sockaddr*)&addr, sizeof addr);

  if ( res == -1){
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
  }

}

/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketListen
 * Signature: (Ljava/io/FileDescriptor;I)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketListen
(JNIEnv *env, jclass cls, jobject fdobj,jint backlog)
{
  int fd,res;
  fd = fdValue(env, fdobj);
  res = listen(fd,backlog);
  if ( res == -1){
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
  }
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    dupFD
 * Signature: (Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;
 */
JNIEXPORT jobject JNICALL Java_hp_infrared_IrSocketImpl_dupFD
(JNIEnv *env, jclass cls, jobject fdobj)
{
  jclass clazz;
  jfieldID id;
  int fd;
  jobject dupObj;
  
  clazz = (*env)->GetObjectClass(env,fdobj);
  id = (*env)->GetFieldID(env,clazz, "fd", "I");
  fd = (*env)->GetIntField(env,fdobj, id);
  dupObj = (*env)->NewObject(env,clazz, (*env)->GetMethodID(env,clazz, 
							    "<init>",
							    "()V"));
  (*env)->SetIntField(env,dupObj, id, fd);
  return dupObj;
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketAccept
 * Signature: (Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;
 */
JNIEXPORT jobject JNICALL Java_hp_infrared_IrSocketImpl_socketAccept
(JNIEnv *env, jclass cls, jobject fdobj, jint timeout, jobject sObj)
{
  jclass clazz;
  jfieldID id;
  int fd,cfd,aret;
  jobject res;
  const char* errmsg;
  struct sockaddr_irda client;
  socklen_t len = sizeof client;
 
  clazz = (*env)->GetObjectClass(env,fdobj);
  id = (*env)->GetFieldID(env,clazz, "fd", "I");
  fd = (*env)->GetIntField(env,fdobj, id);

  if (fd != -1){
    // accept, it may block
    aret = doAccept(fd, timeout, &cfd, (struct sockaddr*)&client, &len,
			    &errmsg);
  }else{
    aret = -1;
    errmsg = "Socket has already been closed.";
  }
  if (aret  == 0){
    res = (*env)->NewObject(env,clazz, (*env)->GetMethodID(env,clazz, "<init>", "()V"));
    (*env)->SetIntField(env,res, id, cfd);
  }else{
    // something wrong
    res = 0;
    if (aret == -1)
      (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), errmsg);
    else 
      // aret == -2, Timeout error
      (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/InterruptedIOException"), errmsg);
  }
  return res;
}

#define DISC_MAX_DEVICES 10
#define DISC_BUF_LEN    sizeof(struct irda_device_list) + \
                        sizeof(struct irda_device_info) * DISC_MAX_DEVICES

/* Perform a discovery on the IrDA link.
 * We don't have a resolver, so we have to look by ourselves and select
 * the device that match the desired nickname/servicename.
 * We could also use auto-connect and let the IrDA stack handle all this
 * for us, it would be a bit nicer and quicker (save a IAS query), but
 * here we can handle errors more properly...
 */
int irda_discover(int fd, const char *nickname, const char *servicename,
		  int *	pdevaddr)
{
  struct irda_device_list * list;           /* List of device */
  unsigned char buf[DISC_BUF_LEN];      /* Actual memory allocation */
  socklen_t len;
  int i,hints,err;
  struct irda_ias_set ias_query;

  /* Set the filter used for performing discovery */
  hints = HINT_COMPUTER | HINT_PDA;
  /* Note : any device may implement socket, so let's not restrict ourselves
   * Jean II */
#if 0
  if (setsockopt(fd, SOL_IRLMP, IRLMP_HINT_MASK_SET, &hints, sizeof(hints))) 
    return(-1);
#endif

  /* Set the list to point to the correct place */
  list = (struct irda_device_list *) buf;
  len = DISC_BUF_LEN;

  /* Ask for the discovery log */
  if(getsockopt(fd, SOL_IRLMP, IRLMP_ENUMDEVICES, buf, &len)) 
      /* Discovery log empty (normal case) or error */
      return(-3);
  /* Is there any addresses ? */
  if(list->len <= 0)
      /* Discovery log empty (exceptional case) */
      return(-3);

  for (i=0; i < list->len; i++)
    {
      /* Check the name if needed */
      if ((strcmp(nickname,"ANY")!=0) &&
	  (strcmp(list->dev[i].info,nickname) != 0))
	continue;

      /* Note : the following bit of code may give you troubles, comment out
       * if needed - Jean II */
      /* Ask if the requested service exist on this device */
      len = sizeof(ias_query);
      ias_query.daddr = list->dev[i].daddr;
      strcpy(ias_query.irda_class_name, servicename);
      strcpy(ias_query.irda_attrib_name, "IrDA:TinyTP:LsapSel");
      err = getsockopt(fd, SOL_IRLMP, IRLMP_IAS_QUERY, &ias_query, &len);

      /* Check for errors */
      if(err == 0)
	{
	  /* No error : Service exist on this device */
	  *pdevaddr = list->dev[i].daddr;
	  return(0);
	}
      else
	{
	  if(errno != EADDRNOTAVAIL)
	    /* Various error codes : Bug in the IrDA stack or link cut */
	    return(-1);
	  else
	    /* Not available : Service not found on device */
	    if(strcmp(nickname,"ANY") != 0)
	      /* We assume only this device has this nickname */
	      return(-2);
	}
    }
  /* End of discovery log... */
  if(strcmp(nickname,"ANY") != 0)
    return(-4);
  else
    return(-2);
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketConnect
 * Signature: (Ljava/io/FileDescriptor;Ljava/lang/String;I)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketConnect
(JNIEnv *env, jclass cls, jobject fdobj, jstring nickname, jint port)
{
  char servicename[20];
  int fd,deviceaddr,res;
  struct sockaddr_irda addr;
  const char *devicename;

  fd = fdValue(env, fdobj);
  if (nickname != NULL) 
	devicename = (*env)->GetStringUTFChars(env,nickname,NULL);
  else devicename = NULL;

  sprintf(servicename,"Service-%d",port);
  deviceaddr = -1;
  res = irda_discover(fd, devicename, servicename, &deviceaddr); 
  if (res == -1) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Discovery error");
    return;
  }
  if (res == -2) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Service not found");
    return;
  }
  if (res == -3) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "No devices in sight");
    return;
  }
  if (res == -4) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Unable to locate the device");
    return;
  }

  addr.sir_family = AF_IRDA;
  sprintf(addr.sir_name,"Service-%d",port);
  addr.sir_addr = deviceaddr;
  // it may block
  res = connect(fd, (struct sockaddr*)&addr, sizeof addr);
  if (res == -1) 
  	(*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketConnectObject
 * Signature: (Ljava/io/FileDescriptor;Ljava/lang/String;Ljava/lang/String;)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketConnectObject
(JNIEnv *env, jclass cls, jobject fdobj, jstring nickname, jstring IASObject)
{
  int fd,deviceaddr,res;
  const char * devicename, *servicename;
  struct sockaddr_irda addr;

  fd = fdValue(env, fdobj);

  if (nickname != NULL) 
	devicename = (*env)->GetStringUTFChars(env,nickname,NULL);
  else devicename = NULL;

  servicename = (*env)->GetStringUTFChars(env,IASObject,NULL);
  deviceaddr = -1;
  res = irda_discover(fd, devicename, servicename, &deviceaddr); 
  if (res == -1) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Discovery error");
    return;
  }
  if (res == -2) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Service not found");
    return;
  }
  if (res == -3) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "No devices in sight");
    return;
  }
  if (res == -4) {
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), "Unable to locate the device");
    return;
  }

  addr.sir_family = AF_IRDA;
  strcpy(addr.sir_name,servicename);
  addr.sir_addr = deviceaddr;

  // it may block
  res = connect(fd, (struct sockaddr*)&addr, sizeof addr);

  if (res == -1) 
	(*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));
}

/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketAvailable
 * Signature: (Ljava/io/FileDescriptor;)I
 */
JNIEXPORT jint JNICALL Java_hp_infrared_IrSocketImpl_socketAvailable
(JNIEnv *env, jclass cls, jobject fdobj)
{
  int fd,bufSize;
  int len = sizeof bufSize;
  unsigned long res = 0;
  int ret;

  fd = fdValue(env, fdobj);
  // not using FIONREAD
  ret = ioctl(fd, TIOCINQ, &res);
  if (ret < 0)
    (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/SocketException"), strerror(errno));

  return (jint)res;
}


/*
 * Class:     hp/infrared/IrSocketImpl
 * Method:    socketClose
 * Signature: (Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;)V
 */
JNIEXPORT void JNICALL Java_hp_infrared_IrSocketImpl_socketClose
(JNIEnv *env, jclass cls, jobject inFdObj, jobject outFdObj)
{
  jclass clazz;
  jfieldID id;
  int fd;

  clazz = (*env)->GetObjectClass(env,inFdObj);
  id = (*env)->GetFieldID(env,clazz, "fd", "I");
  fd = (*env)->GetIntField(env,inFdObj, id);

  if (fd != -1){
    int res;

    // Change fds to invalid fd
    (*env)->SetIntField(env,inFdObj, id, (jint)-1);
    if (outFdObj)
      (*env)->SetIntField(env,outFdObj, id, (jint)-1);

    res = close(fd);
    if (res < 0 && errno != EBADF)
      (*env)->ThrowNew(env,(*env)->FindClass(env,"java/io/IOException"), strerror(errno));

  }
}
